from typing import Tuple
import numpy as np
import numpy.typing as npt
import spikeinterface as si


def get_times_labels_from_sorting(sorting: si.BaseSorting) -> Tuple[npt.NDArray[np.int64], npt.NDArray[np.int32]]:
    """Get times and labels from a sorting object
    Inputs:
        sorting: a sorting object
    Returns:
        times: 1D array of spike times
        labels: 1D array of spike labels
    Example:
        times, labels = get_times_labels_from_sorting(sorting)
    """
    times_list = []
    labels_list = []
    for unit_id in sorting.get_unit_ids():
        times0 = sorting.get_unit_spike_train(unit_id=unit_id)
        labels0 = np.ones(times0.shape, dtype=np.int32) * unit_id
        times_list.append(times0.astype(np.int64))
        labels_list.append(labels0)
    if len(times_list) > 0:
        times = np.concatenate(times_list).astype(np.int64)
        labels = np.concatenate(labels_list)
        inds = np.argsort(times)
        times = times[inds]
        labels = labels[inds]
        return times, labels
    else:
        return np.array([], dtype=np.int64), np.array([], dtype=np.int32)


from typing import Dict, Union
import numpy as np
import numpy.typing as npt
import spikeinterface as si
from .Scheme3SortingParameters import Scheme3SortingParameters
from .sorting_scheme2 import get_time_chunks
from .sorting_scheme2 import sorting_scheme2, get_times_labels_from_sorting
from ..core.get_block_recording_for_scheme3 import get_block_recording_for_scheme3
from ..core.SnippetClassifier import SnippetClassifier
from ..core.get_times_labels_from_sorting import get_times_labels_from_sorting


def sorting_scheme3(
    recording: si.BaseRecording, *,
    sorting_parameters: Scheme3SortingParameters
) -> si.BaseSorting:
    """MountainSort 5 sorting scheme 3

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        sorting_parameters (Scheme3SortingParameters): Sorting parameters

    Returns:
        si.BaseSorting: SpikeInterface sorting object
    """

    if recording.get_num_segments() > 1:
        print('Recording has multiple segments. Joining segments for sorting...')
        recording_joined = si.concatenate_recordings(recording_list=[recording])
        sorting_joined = sorting_scheme3(recording_joined, sorting_parameters=sorting_parameters)
        print('Splitting sorting into segments to match original multisegment recording...')
        sorting = si.split_sorting(sorting_joined, recording_joined)
        return sorting

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency
    channel_locations = recording.get_channel_locations()

    sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)

    block_size = int(sorting_parameters.block_duration_sec * sampling_frequency) # size of chunks in samples
    blocks = get_time_chunks(recording.get_num_samples(), chunk_size=block_size, padding=1000)

    times_list: list[npt.NDArray[np.int64]] = []
    labels_list: list[npt.NDArray[np.int32]] = []
    last_label_used = 0
    previous_snippet_classifiers: Union[Dict[int, SnippetClassifier], None] = None
    for i, chunk in enumerate(blocks):
        print('')
        print('=============================================')
        print(f'Processing block {i + 1} of {len(blocks)}...')
        subrecording = get_block_recording_for_scheme3(recording=recording, start_frame=chunk.start - chunk.padding_left, end_frame=chunk.end + chunk.padding_right)
        subsorting, snippet_classifiers = sorting_scheme2(
            subrecording,
            sorting_parameters=sorting_parameters.block_sorting_parameters,
            return_snippet_classifiers=True,
            reference_snippet_classifiers=previous_snippet_classifiers,
            label_offset=last_label_used
        )
        previous_snippet_classifiers = snippet_classifiers
        times0, labels0 = get_times_labels_from_sorting(subsorting)
        valid_inds = np.where((times0 >= chunk.padding_left) & (times0 < chunk.padding_left + (chunk.end - chunk.start)))[0]
        times0: npt.NDArray[np.int32] = times0[valid_inds]
        labels0: npt.NDArray[np.int32] = labels0[valid_inds]
        times0 = times0.astype(np.int64) + chunk.start - np.int64(chunk.padding_left)
        if len(labels0) > 0:
            last_label_used = max(last_label_used, np.max(labels0))
        times_list.append(times0)
        labels_list.append(labels0)
    
    times_concat = np.concatenate(times_list)
    labels_concat = np.concatenate(labels_list)

    # Now create a new sorting object from the times and labels results
    sorting2 = si.NumpySorting.from_times_labels([times_concat], [labels_concat], sampling_frequency=recording.sampling_frequency)

    return sorting2


from typing import List, Tuple
import numpy as np
import numpy.typing as npt
from .compute_pca_features import compute_pca_features
from isosplit6 import isosplit6


def pairwise_merge_step(*, snippets: npt.NDArray[np.float32], templates: npt.NDArray[np.float32], labels: npt.NDArray[np.int32], times: npt.NDArray[np.int32], detect_sign: int, unit_ids: np.array, detect_time_radius: int) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.int32]]:
    # L = snippets.shape[0]
    # T = snippets.shape[1]
    # M = snippets.shape[2]
    K = templates.shape[0]

    new_times = np.array(times)
    new_labels = np.array(labels)

    if detect_sign < 0:
        A = -templates
    elif detect_sign > 0: # pragma: no cover
        A = templates # pragma: no cover
    else:
        A = np.abs(templates) # pragma: no cover
    
    merges = {}
    peak_channel_indices = [np.argmax(np.max(A[i], axis=0)) for i in range(K)]
    peak_values_on_channels = [np.max(A[i], axis=0) for i in range(K)]
    peak_time_indices_on_channels = [np.argmax(A[i], axis=0) for i in range(K)]
    for i1 in range(K):
        for i2 in range(i1 - 1, -1, -1): # merge everything into lower ID - important to go in reverse order
            offset1 = peak_time_indices_on_channels[i2][peak_channel_indices[i1]] - peak_time_indices_on_channels[i1][peak_channel_indices[i1]]
            offset2 = peak_time_indices_on_channels[i2][peak_channel_indices[i2]] - peak_time_indices_on_channels[i1][peak_channel_indices[i2]]
            if np.abs(offset1 - offset2) <= 4: # make sure the offsets are roughly consistent
                if peak_values_on_channels[i1][peak_channel_indices[i2]] > 0.5 * peak_values_on_channels[i2][peak_channel_indices[i2]]:
                    if peak_values_on_channels[i2][peak_channel_indices[i1]] > 0.5 * peak_values_on_channels[i1][peak_channel_indices[i1]]:
                        # print(f'Pairwise merge: comparing units {unit_ids[i1]} and {unit_ids[i2]} (offset: {offset1})')
                        if test_merge(offset_snippets(snippets[new_labels == unit_ids[i1]], offset=offset1), snippets[new_labels == unit_ids[i2]]):
                            print(f'Pairwise merge: ** merging {unit_ids[i1]} and {unit_ids[i2]} (offset: {offset1})')
                            merges[unit_ids[i1]] = {
                                'unit_id1': unit_ids[i1],
                                'unit_id2': unit_ids[i2],
                                'offset': offset1
                            }
    for unit_id in unit_ids[::-1]: # important to go in descending order for transitive merges to work
        if unit_id in merges:
            mm = merges[unit_id]
            unit_id1 = mm['unit_id1']
            unit_id2 = mm['unit_id2']
            assert unit_id == unit_id1
            offset = mm['offset']
            inds1 = np.nonzero(new_labels == unit_id1)[0]
            inds2 = np.nonzero(new_labels == unit_id2)[0]
            print(f'Performing merge of units {unit_id1} ({len(inds1)} events) and {unit_id2} ({len(inds2)} events) (offset: {offset})')
            new_times[inds1] = new_times[inds1] + offset
            new_labels[inds1] = unit_id2
    
    # need to re-sort the times after the merges (because of the time offsets)
    sort_inds = np.argsort(new_times)
    new_times = new_times[sort_inds]
    new_labels = new_labels[sort_inds]
    
    # After merges, we may now have duplicate events - let's remove them
    new_inds = remove_duplicate_events(new_times, new_labels, tol=detect_time_radius)
    print(f'Removing {len(times) - len(new_inds)} duplicate events')
    new_times = new_times[new_inds]
    new_labels = new_labels[new_inds]

    return new_times, new_labels

def remove_duplicate_events(times: npt.NDArray[np.int32], labels: npt.NDArray[np.int32], *, tol: int) -> npt.NDArray[np.int32]:
    new_labels = np.array(labels)
    unit_ids = np.unique(new_labels)
    for unit_id in unit_ids:
        unit_inds = np.nonzero(new_labels == unit_id)[0]
        unit_times = times[unit_inds]
        inds_duplicate = find_duplicate_times(unit_times, tol=tol)
        new_labels[unit_inds[inds_duplicate]] = 0
    inds_nonzero = np.nonzero(new_labels)[0]
    return inds_nonzero

def find_duplicate_times(times: npt.NDArray[np.int32], *, tol: int) -> npt.NDArray[np.int32]:
    ret: list[np.int32] = []
    deleted = np.zeros((len(times),), dtype=np.int16)
    for i1 in range(len(times)):
        if not deleted[i1]:
            i2 = i1 + 1
            while i2 < len(times) and times[i2] <= times[i1] + tol:
                ret.append(i2)
                deleted[i2] = True
                i2 += 1
    return np.array(ret, dtype=np.int32)

def test_merge(snippets1: npt.NDArray[np.float32], snippets2: npt.NDArray[np.float32]) -> bool:
    L1 = snippets1.shape[0]
    L2 = snippets2.shape[0]
    T = snippets1.shape[1]
    M = snippets1.shape[2]
    V1 = snippets1.reshape((L1, T * M))
    V2 = snippets2.reshape((L2, T * M))
    Vall = np.concatenate([V1, V2], axis=0)
    features0 = compute_pca_features(Vall, npca=12) # should we hard-code this as 12?
    labels0 = isosplit6(features0)
    # not sure best way to test this, but for now we'll check that we only have single cluster after isosplit
    # after all, each cluster should not split on its own (branch cluster method)
    return np.max(labels0) == 1
    # dominant_label1 = np.argmax(np.bincount(labels0[:L1]))
    # dominant_label2 = np.argmax(np.bincount(labels0[L1:]))
    # return dominant_label1 == dominant_label2

def offset_snippets(snippets: npt.NDArray[np.float32], *, offset: int):
    return np.roll(snippets, shift=offset, axis=1)


import time
import spikeinterface as si
import spikeinterface.extractors as se
import spikeinterface.preprocessing as spre
import spikeinterface.comparison as sc
import mountainsort5 as ms5


def test_scheme2():
    recording, sorting_true = se.toy_example(
        duration=40,
        num_channels=8,
        num_units=16,
        sampling_frequency=30000,
        num_segments=2,
        seed=0
    )

    # lazy preprocessing
    recording_filtered = spre.bandpass_filter(recording, freq_min=300, freq_max=6000)
    recording_preprocessed: si.BaseRecording = spre.whiten(recording_filtered)

    # sorting
    print('Starting MountainSort5 (sorting1)')
    timer = time.time()
    sorting1 = ms5.sorting_scheme2(
        recording_preprocessed,
        sorting_parameters=ms5.Scheme2SortingParameters(
            phase1_detect_channel_radius=150,
            detect_channel_radius=50,
            max_num_snippets_per_training_batch=3, # for improving test coverage
            snippet_mask_radius=150,
            training_duration_sec=15
        )
    )
    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    print('Starting MountainSort5 (sorting2)')
    timer = time.time()
    sorting2 = ms5.sorting_scheme2(
        recording_preprocessed,
        sorting_parameters=ms5.Scheme2SortingParameters(
            phase1_detect_channel_radius=150,
            detect_channel_radius=50,
            training_duration_sec=25,
            training_recording_sampling_mode='uniform'
        )
    )
    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    # commenting out because this step takes a while
    # print('Comparing with truth')
    # comparison = sc.compare_sorter_to_ground_truth(gt_sorting=sorting_true, tested_sorting=sorting)
    # print(comparison.get_performance())


import numpy as np
import numpy.typing as npt
from sklearn import decomposition


def compute_pca_features(X: npt.NDArray[np.float32], *, npca: int):
    pca = decomposition.PCA(n_components=np.minimum(npca, X.shape[0]))
    return pca.fit_transform(X)


import time
import spikeinterface as si
import spikeinterface.extractors as se
import spikeinterface.preprocessing as spre
import spikeinterface.comparison as sc
import mountainsort5 as ms5


def test_scheme1():
    recording, sorting_true = se.toy_example(
        duration=20,
        num_channels=8,
        num_units=16,
        sampling_frequency=30000,
        num_segments=2,
        seed=0
    )

    timer = time.time()

    # lazy preprocessing
    recording_filtered = spre.bandpass_filter(recording, freq_min=300, freq_max=6000)
    recording_preprocessed: si.BaseRecording = spre.whiten(recording_filtered)

    # sorting
    print('Starting MountainSort5')
    sorting = ms5.sorting_scheme1(
        recording_preprocessed,
        sorting_parameters=ms5.Scheme1SortingParameters(
            snippet_mask_radius=50
        )
    )
    
    elapsed_sec = time.time() - timer
    duration_sec = recording.get_total_duration()
    print(f'Elapsed time for sorting: {elapsed_sec:.2f} sec -- x{(duration_sec / elapsed_sec):.2f} speed compared with real time for {recording.get_num_channels()} channels')

    # commenting out because this step takes a while
    # print('Comparing with truth')
    # comparison = sc.compare_sorter_to_ground_truth(gt_sorting=sorting_true, tested_sorting=sorting)
    # print(comparison.get_performance())


import numpy as np
from typing import Union, Literal
from dataclasses import dataclass
from .Scheme2SortingParameters import Scheme2SortingParameters


@dataclass
class Scheme3SortingParameters:
    """Parameters for MountainSort sorting scheme 3

    - block_sorting_parameters: Scheme2SortingParameters for individual blocks
    - block_duration_sec: duration of each block
    """
    block_sorting_parameters: Scheme2SortingParameters
    block_duration_sec: float

    def check_valid(self, *, M: int, N: int, sampling_frequency: float, channel_locations: Union[np.ndarray, None]=None):
        """Internal function for checking validity of parameters"""
        self.block_sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)
        assert self.block_duration_sec > 0



from typing import Union, List
import numpy as np
import numpy.typing as npt


def extract_snippets(
    traces: npt.NDArray[np.float32], *,
    channel_locations: Union[npt.NDArray[np.float32], None],
    mask_radius: Union[float, None],
    times: npt.NDArray[np.int32],
    channel_indices: Union[npt.NDArray[np.int32], None],
    T1: int,
    T2: int
) -> npt.NDArray[np.float32]:
    M = traces.shape[1]
    L = len(times)

    if mask_radius is not None:
        assert channel_locations is not None
        assert channel_indices is not None
        adjacency = []
        for m in range(M):
            adjacency.append([])
            for m2 in range(M):
                dist0 = np.sqrt(np.sum((channel_locations[m] - channel_locations[m2]) ** 2))
                if dist0 <= mask_radius:
                    adjacency[m].append(m2)
    else:
        adjacency = None
    
    snippets = np.zeros((L, T1 + T2, M), dtype=np.float32)
    for j in range(L):
        t1 = times[j] - T1
        t2 = times[j] + T2
        if adjacency is not None:
            assert channel_indices is not None
            channel_inds = adjacency[channel_indices[j]]
            snippets[j][:, channel_inds] = traces[t1:t2, channel_inds]
        else:
            snippets[j] = traces[t1:t2]
    return snippets

def extract_snippets_in_channel_neighborhood(
    traces: npt.NDArray[np.float32], *,
    times: npt.NDArray[np.int32],
    neighborhood: Union[List[int], None],
    T1: int,
    T2: int
) -> np.ndarray:
    L = len(times)

    if neighborhood is None:
        neighborhood = list(range(traces.shape[1]))

    snippets = np.zeros((L, T1 + T2, len(neighborhood)), dtype=np.float32)
    for j in range(L):
        t1 = times[j] - T1
        t2 = times[j] + T2
        snippets[j] = traces[t1:t2][:, neighborhood]
            
    return snippets


from typing import Tuple, Union
import numpy as np
import numpy.typing as npt


def detect_spikes(
    traces: npt.NDArray[np.float32], *,
    channel_locations: npt.NDArray[np.float32],
    time_radius: int,
    channel_radius: Union[float, None],
    detect_threshold: float,
    detect_sign: int,
    margin_left: int,
    margin_right: int,
    verbose: bool
) -> Tuple[npt.NDArray[np.int32], npt.NDArray[np.int32]]:
    N = traces.shape[0]
    M = traces.shape[1]

    if detect_sign > 0:
        # todo: figure out how to avoid making a copy
        traces = -traces # pragma: no cover
    elif detect_sign == 0:
        # todo: figure out how to avoid making a copy
        traces = -np.abs(traces) # pragma: no cover
    
    adjacency = []
    for m in range(M):
        adjacency.append([])
        for m2 in range(M):
            dist0 = np.sqrt(np.sum((channel_locations[m] - channel_locations[m2]) ** 2))
            if (channel_radius is None) or (dist0 <= channel_radius):
                adjacency[m].append(m2)
    
    inds1, inds2 = np.nonzero(traces <= -detect_threshold)
    
    candidate_times = [[] for m in range(M)]
    candidate_values = [[] for m in range(M)]
    for i in range(len(inds1)):
        if inds1[i] >= margin_left and inds1[i] < N - margin_right:
            candidate_times[inds2[i]].append(inds1[i])
            candidate_values[inds2[i]].append(traces[inds1[i], inds2[i]])

    times = []
    channel_indices = []
    for m in range(M):
        nbhd = adjacency[m]
        if verbose:
            print(f'm = {m} (nbhd size: {len(nbhd)})')
        indices = [0 for j in range(len(nbhd))]
        for i in range(len(candidate_times[m])):
            t = candidate_times[m][i]
            v = candidate_values[m][i]
            okay = True
            for j in range(len(nbhd)):
                if not okay:
                    break
                tt = candidate_times[nbhd[j]]
                vv = candidate_values[nbhd[j]]
                ii = indices[j]
                while ii < len(tt) and tt[ii] < t - time_radius:
                    ii += 1
                indices[j] = ii # advance
                jj = ii
                while jj < len(tt) and tt[jj] <= t + time_radius:
                    if vv[jj] < v:
                        okay =False
                        break
                    jj += 1
            if okay:
                times.append(t)
                channel_indices.append(m)
    
    times = np.array(times, dtype=np.int32)
    channel_indices = np.array(channel_indices, dtype=np.int32)
    inds = np.argsort(times)
    times = times[inds]
    channel_indices = channel_indices[inds]
    return times, channel_indices




from typing import Union
import numpy as np
import numpy.typing as npt
from isosplit6 import isosplit6
from .compute_pca_features import compute_pca_features


def cluster_snippets(
    snippets: npt.NDArray[np.float32], *,
    npca_per_branch: int,
    snippet_inds: Union[npt.NDArray[np.int32], None]=None # pass in inds so that we don't keep making copies of the array
) -> npt.NDArray[np.int32]:
    if snippet_inds is not None:
        snippets_sub = snippets[snippet_inds]
    else:
        snippets_sub = snippets
    L = snippets_sub.shape[0]
    T = snippets_sub.shape[1]
    M = snippets_sub.shape[2]
    features = compute_pca_features(snippets_sub.reshape((L, T * M)), npca=npca_per_branch)
    snippets_sub = None # free up memory
    labels = isosplit6(features)
    K = np.max(labels)
    if K <= 1:
        return labels
    print(f'{K} clusters in branch')
    ret = np.zeros((L,), dtype=np.int32)
    last_ret_k = 0
    for k in range(1, K + 1):
        inds0 = np.nonzero(labels == k)[0]
        if snippet_inds is not None:
            inds1 = snippet_inds[inds0]
        else:
            inds1 = inds0
        labels0 = cluster_snippets(snippets, npca_per_branch=npca_per_branch, snippet_inds=inds1)
        K2 = np.max(labels0)
        for k2 in range(1, K2 + 1):
            last_ret_k += 1
            ret[inds0[labels0 == k2]] = last_ret_k
    return ret


from typing import Union
import numpy as np
import math
import spikeinterface as si
from .Scheme1SortingParameters import Scheme1SortingParameters
from ..core.detect_spikes import detect_spikes
from ..core.extract_snippets import extract_snippets
from ..core.cluster_snippets import cluster_snippets
from ..core.compute_templates import compute_templates
from ..core.pairwise_merge_step import pairwise_merge_step


def sorting_scheme1(
    recording: si.BaseRecording, *,
    sorting_parameters: Scheme1SortingParameters
):
    """MountainSort 5 sorting scheme 1

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        sorting_parameters (Scheme2SortingParameters): Sorting parameters

    Returns:
        si.BaseSorting: SpikeInterface sorting object
    """

    if recording.get_num_segments() > 1:
        print('Recording has multiple segments. Joining segments for sorting...')
        recording_joined = si.concatenate_recordings(recording_list=[recording])
        sorting_joined = sorting_scheme1(recording_joined, sorting_parameters=sorting_parameters)
        print('Splitting sorting into segments to match original multisegment recording...')
        sorting = si.split_sorting(sorting_joined, recording_joined)
        return sorting

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency

    channel_locations = recording.get_channel_locations()

    print(f'Number of channels: {M}')
    print(f'Number of timepoints: {N}')
    print(f'Sampling frequency: {sampling_frequency} Hz')
    for m in range(M):
        print(f'Channel {m}: {channel_locations[m]}')

    sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)
    
    print('Loading traces')
    traces = recording.get_traces()

    print('Detecting spikes')
    time_radius = int(math.ceil(sorting_parameters.detect_time_radius_msec / 1000 * sampling_frequency))
    times, channel_indices = detect_spikes(
        traces=traces,
        channel_locations=channel_locations,
        time_radius=time_radius,
        channel_radius=sorting_parameters.detect_channel_radius,
        detect_threshold=sorting_parameters.detect_threshold,
        detect_sign=sorting_parameters.detect_sign,
        margin_left=sorting_parameters.snippet_T1,
        margin_right=sorting_parameters.snippet_T2,
        verbose=True
    )

    print(f'Extracting {len(times)} snippets')
    snippets = extract_snippets( # L x T x M
        traces=traces,
        channel_locations=channel_locations,
        mask_radius=sorting_parameters.snippet_mask_radius,
        times=times,
        channel_indices=channel_indices,
        T1=sorting_parameters.snippet_T1,
        T2=sorting_parameters.snippet_T2
    )
    L = snippets.shape[0]
    T = snippets.shape[1]
    assert snippets.shape[2] == M

    print('Clustering snippets')
    labels = cluster_snippets(
        snippets=snippets,
        npca_per_branch=sorting_parameters.npca_per_branch
    )
    K = int(np.max(labels))
    print(f'Found {K} clusters')

    print('Computing templates')
    templates = compute_templates(snippets=snippets, labels=labels) # K x T x M
    peak_channel_indices = [np.argmin(np.min(templates[i], axis=0)) for i in range(K)]

    if sorting_parameters.pairwise_merge_step:
        print('Pairwise merge step')
        times, labels = pairwise_merge_step(
            templates=templates,
            snippets=snippets,
            labels=labels,
            times=times,
            detect_sign=sorting_parameters.detect_sign,
            unit_ids=np.arange(1, K + 1).astype(np.int32),
            detect_time_radius=time_radius
        )

    print('Reordering units')
    # relabel so that units are ordered by channel
    # and we also put any labels that are not used at the end
    aa = peak_channel_indices
    for k in range(1, K + 1):
        inds = np.where(labels == k)[0]
        if len(inds) == 0:
            aa[k - 1] = np.Inf
    new_labels_mapping = np.argsort(np.argsort(aa)) + 1 # too tricky! my head aches right now
    labels = new_labels_mapping[labels - 1]
    
    sorting = si.NumpySorting.from_times_labels(times_list=[times], labels_list=[labels], sampling_frequency=sampling_frequency)

    return sorting


import numpy as np
import numpy.typing as npt
from typing import Union, Literal
from dataclasses import dataclass


@dataclass
class Scheme2SortingParameters:
    """Parameters for MountainSort sorting scheme 2

    See Scheme1SortingParameters for more details on the parameters below.

    - phase1_detect_channel_radius: detect_channel_radius in phase 1
    - detect_channel_radius: detect_channel_radius in phase 2
    - phase1_detect_threshold: detect_threshold in phase 1
    - phase1_detect_time_radius_msec: detect_time_radius_msec in phase 1
    - detect_time_radius_msec: detect_time_radius_msec in phase 2
    - phase1_npca_per_branch: npca_per_branch in phase 1
    - phase1_pairwise_merge_step: pairwise_merge_step in phase 1
    - detect_sign
    - detect_threshold: detect_threshold in phase 2
    - snippet_T1
    - snippet_T2
    - snippet_mask_radius
    - max_num_snippets_per_training_batch: the maximum number of snippets to use for training the classifier in each batch
    - classifier_npca: the number of principal components to use for the classifier
    - training_duration_sec: the duration of the training data (in seconds)
    - training_recording_sampling_mode: how to sample the training data. If 'initial', then the first training_duration_sec of the recording will be used. If 'uniform', then the training data will be sampled uniformly in 10-second chunks from the recording.
    """
    phase1_detect_channel_radius: Union[float, None]
    detect_channel_radius: Union[float, None]
    phase1_detect_threshold: float=5.5
    phase1_detect_time_radius_msec: float=1.5
    detect_time_radius_msec: float=0.5
    phase1_npca_per_branch: int=12
    phase1_pairwise_merge_step: bool=True
    detect_sign: int=-1
    detect_threshold: float=5.5
    snippet_T1: int=20
    snippet_T2: int=20
    snippet_mask_radius: Union[float, None]=None
    max_num_snippets_per_training_batch: int=200
    classifier_npca: Union[int, None]=60
    training_duration_sec: Union[float, None]=None
    training_recording_sampling_mode: Literal['initial', 'uniform']='initial'

    def check_valid(self, *, M: int, N: int, sampling_frequency: float, channel_locations: npt.NDArray[np.float32]):
        """Internal function for checking validity of parameters"""
        assert channel_locations.shape[0] == M, 'Shape mismatch between traces and channel locations'
        D = channel_locations.shape[1]
        assert N >= self.snippet_T1 + self.snippet_T2
        if self.snippet_mask_radius is not None:
            assert self.snippet_mask_radius >= 0
        assert M >= 1 and M < 1e6
        assert D >= 1 and D <= 3
        assert sampling_frequency > 0 and sampling_frequency <= 1e7
        if self.phase1_detect_channel_radius is not None:
            assert self.phase1_detect_channel_radius > 0
        assert self.phase1_detect_time_radius_msec > 0 and self.phase1_detect_time_radius_msec <= 1e4
        assert self.phase1_detect_threshold > 0
        assert self.detect_sign in [-1, 0, 1]
        assert self.phase1_npca_per_branch >= 1 and self.phase1_npca_per_branch <= 1e3




import spikeinterface as si
from typing import Literal


def get_block_recording_for_scheme3(
    recording: si.BaseRecording, *,
    start_frame: int,
    end_frame: int
) -> si.BaseRecording:
    return BlockRecording(
        recording=recording,
        start_frame=start_frame,
        end_frame=end_frame
    )

class BlockRecording(si.BaseRecording):
    def __init__(self, recording: si.BaseRecording, start_frame: int, end_frame: int):
        sampling_frequency = recording.get_sampling_frequency()
        dtype = recording.get_dtype()
        channel_ids = recording.get_channel_ids()
        
        si.BaseRecording.__init__(self, sampling_frequency, channel_ids, dtype)

        self.start_frame = start_frame
        self.end_frame = end_frame

        self.set_channel_locations(recording.get_channel_locations())
        self.is_dumpable = False

        self.add_recording_segment(
            BlockRecordingSegment(recording=recording, start_frame=start_frame, end_frame=end_frame)
        )

        self._kwargs = {'recording': recording, 'start_frame': start_frame, 'end_frame': end_frame}

class BlockRecordingSegment(si.BaseRecordingSegment):
    def __init__(self, recording: si.BaseRecording, start_frame: int, end_frame: int):
        si.BaseRecordingSegment.__init__(
            self,
            sampling_frequency=recording.get_sampling_frequency(),
            t_start=0
        )
        self._recording = recording
        self._start_frame = start_frame
        self._end_frame = end_frame

    def get_num_samples(self):
        return self._end_frame - self._start_frame

    def get_traces(self, start_frame, end_frame, channel_indices):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_samples()

        # Get the traces from the parent recording
        return self._recording._recording_segments[0].get_traces(
            start_frame=start_frame + self._start_frame,
            end_frame=end_frame + self._start_frame,
            channel_indices=channel_indices
        )


import numpy as np
import numpy.typing as npt
from typing import Union
from dataclasses import dataclass


@dataclass
class Scheme1SortingParameters:
    """Parameters for MountainSort sorting scheme 1
    - detect_threshold: the threshold for detection of whitened data
    - detect_channel_radius: the radius (in units of channel locations) for exluding nearby channels from detection
    - detect_time_radius_msec: the radius (in msec) for excluding nearby events from detection
    - detect_sign: the sign of the threshold for detection (1, -1, or 0)
    - snippet_T1: the number of timepoints before the event to include in the snippet
    - snippet_T2: the number of timepoints after the event to include in the snippet
    - snippet_mask_radius: the radius (in units of channel locations) for making a snippet around the central channel
    - npca_per_branch: the number of PCA components to compute for each branch of clustering
    - pairwise_merge_step: whether to do the pairwise merge step
    """
    detect_threshold: float=5.5
    detect_channel_radius: Union[float, None]=None
    detect_time_radius_msec: float=0.5
    detect_sign: int=-1
    snippet_T1: int=20
    snippet_T2: int=20
    snippet_mask_radius: Union[float, None]=None
    npca_per_branch: int=12
    pairwise_merge_step: bool=True

    def check_valid(self, *, M: int, N: int, sampling_frequency: float, channel_locations: npt.NDArray[np.float32]):
        """Internal function for checking validity of parameters"""
        assert channel_locations.shape[0] == M, 'Shape mismatch between traces and channel locations'
        D = channel_locations.shape[1]
        assert N >= self.snippet_T1 + self.snippet_T2
        if self.snippet_mask_radius is not None:
            assert self.snippet_mask_radius >= 0
        assert M >= 1 and M < 1e6
        assert D >= 1 and D <= 3
        assert sampling_frequency > 0 and sampling_frequency <= 1e7
        if self.detect_channel_radius is not None:
            assert self.detect_channel_radius > 0
        assert self.detect_time_radius_msec > 0 and self.detect_time_radius_msec <= 1e4
        assert self.detect_threshold > 0
        assert self.detect_sign in [-1, 0, 1]
        assert self.npca_per_branch >= 1 and self.npca_per_branch <= 1e3




import numpy as np
import spikeinterface as si
import spikeinterface.extractors as se
from mountainsort5.core.compute_pca_features import compute_pca_features
from mountainsort5.core.compute_templates import compute_templates
from mountainsort5.core.detect_spikes import detect_spikes
from mountainsort5.core.extract_snippets import extract_snippets, extract_snippets_in_channel_neighborhood
from mountainsort5.core.get_sampled_recording_for_training import get_sampled_recording_for_training
from mountainsort5.core.get_block_recording_for_scheme3 import get_block_recording_for_scheme3
from mountainsort5.core.cluster_snippets import cluster_snippets
from mountainsort5.core.pairwise_merge_step import pairwise_merge_step
from mountainsort5.core.get_times_labels_from_sorting import get_times_labels_from_sorting


def test_compute_pca_features():
    x = np.random.normal(size=(100, 20))
    features = compute_pca_features(x, npca=10)
    assert features.shape == (100, 10)

def test_compute_templates():
    L = 1000
    T = 20
    M = 10
    snippets = np.random.normal(size=(L, T, M))
    labels = np.random.randint(1, 11, size=(L,))
    templates = compute_templates(snippets, labels)
    assert templates.shape == (10, T, M)

def test_detect_spikes():
    N = 100
    M = 3
    traces= np.zeros((N, M))
    channel_locations = np.zeros((M, 2))
    traces[8, 0] = -10 # in margin
    traces[9, 1] = -4 # in margin
    traces[30, 1] = -4
    traces[31, 1] = -10
    traces[40, 2] = -10
    traces[41, 2] = -4
    traces[98, 0] = -10 # in margin
    times, channel_indices = detect_spikes(
        traces,
        channel_locations=channel_locations,
        time_radius=10,
        channel_radius=50,
        detect_threshold=3,
        detect_sign=-1,
        margin_left=10,
        margin_right=10,
        verbose=False
    )
    assert len(times) == len(channel_indices) == 2
    assert times[0] == 31
    assert channel_indices[0] == 1
    assert times[1] == 40
    assert channel_indices[1] == 2

def test_extract_snippets():
    N = 1000
    M = 4
    L = 100
    T1 = 20
    T2 = 20
    T = T1 + T2
    traces = np.random.normal(size=(N, M))
    times = np.random.randint(T1, N - T2, size=(L,))
    snippets = extract_snippets(traces, times=times, channel_locations=None, mask_radius=None, channel_indices=None, T1=T1, T2=T2)
    assert snippets.shape == (L, T, M)

def test_extract_snippets_in_channel_neighborhood():
    N = 1000
    M = 4
    L = 100
    T1 = 20
    T2 = 20
    T = T1 + T2
    traces = np.random.normal(size=(N, M))
    times = np.random.randint(T1, N - T2, size=(L,))
    neighborhood = [0, 2]
    snippets = extract_snippets_in_channel_neighborhood(
        traces=traces,
        times=times,
        neighborhood=neighborhood,
        T1=T1,
        T2=T2
    )
    assert snippets.shape == (L, T, len(neighborhood))

def test_get_sampled_recording_for_training():
    recording, _ = se.toy_example(duration=60, num_channels=4, num_units=10, sampling_frequency=30000, seed=0, num_segments=1)
    recording: si.BaseRecording = recording

    # test case where the training duration is longer than the recording duration
    recording1 = get_sampled_recording_for_training(
        recording=recording,
        training_duration_sec=100,
        mode='initial'
    )
    assert np.array_equal(
        recording1.get_traces(),
        recording.get_traces()
    )

    # test case where the training recording has only one chunk
    recording1b = get_sampled_recording_for_training(
        recording=recording,
        training_duration_sec=7,
        mode='uniform'
    )
    assert np.array_equal(
        recording1b.get_traces(),
        recording.get_traces(start_frame=0, end_frame=7 * recording.sampling_frequency)
    )

    # Test the mode='initial' case which is easier
    recording2 = get_sampled_recording_for_training(
        recording=recording,
        training_duration_sec=25,
        mode='initial'
    )
    assert recording2.get_num_frames() == 25 * recording.sampling_frequency
    assert np.array_equal(
        recording2.get_traces(),
        recording.get_traces(start_frame=0, end_frame=25 * recording.sampling_frequency)
    )

    # Important to test mode='uniform' which can be tricky
    recording3 = get_sampled_recording_for_training(
        recording=recording,
        training_duration_sec=25,
        mode='uniform'
    )
    assert recording3.get_num_frames() == 25 * recording.sampling_frequency
    assert np.array_equal(
        recording3.get_traces(start_frame=0, end_frame=10 * recording.sampling_frequency),
        recording.get_traces(start_frame=0, end_frame=10 * recording.sampling_frequency)
    )
    # 60 - 25 = 35
    # spacing = 35 / 2 = 17.5
    assert np.array_equal(
        recording3.get_traces(start_frame=10 * recording.sampling_frequency, end_frame=20 * recording.sampling_frequency),
        recording.get_traces(start_frame=int(27.5 * recording.sampling_frequency), end_frame=int(37.5 * recording.sampling_frequency))
    )
    # 37.5 + 17.5 = 55
    assert np.array_equal(
        recording3.get_traces(start_frame=20 * recording.sampling_frequency, end_frame=25 * recording.sampling_frequency),
        recording.get_traces(start_frame=int(55 * recording.sampling_frequency), end_frame=int(60 * recording.sampling_frequency))
    )

def test_get_block_recording_for_scheme3():
    recording, _ = se.toy_example(duration=60, num_channels=4, num_units=10, sampling_frequency=30000, seed=0, num_segments=1)
    recording: si.BaseRecording = recording

    recording2 = get_block_recording_for_scheme3(
        recording=recording,
        start_frame=10 * recording.sampling_frequency,
        end_frame=20 * recording.sampling_frequency
    )
    assert np.array_equal(
        recording.get_traces(start_frame=10 * recording.sampling_frequency, end_frame=20 * recording.sampling_frequency),
        recording2.get_traces()
    )

def test_cluster_snippets():
    N = 1000
    L = 100
    M = 4
    T1 = 20
    T2 = 20
    traces = np.random.normal(size=(N, M))
    traces[500:] += 10 # offset this so we get more than one cluster (important for coverage of cluster_snippets)
    times = np.random.randint(T1, N - T2, size=(L,))
    snippets = extract_snippets(traces, times=times, channel_locations=None, mask_radius=None, channel_indices=None, T1=T1, T2=T2)
    labels = cluster_snippets(
        snippets,
        npca_per_branch=10
    )
    assert np.min(labels) == 1
    assert len(labels) == L

def test_pairwise_merge_step():
    N = 1000
    L = 100
    M = 4
    T1 = 20
    T2 = 20
    traces = np.random.normal(size=(N, M))
    times = np.random.randint(T1, N - T2, size=(L,))
    snippets = extract_snippets(traces, times=times, channel_locations=None, mask_radius=None, channel_indices=None, T1=T1, T2=T2)
    labels = cluster_snippets(
        snippets,
        npca_per_branch=10
    )
    templates = compute_templates(snippets, labels)
    new_times, new_labels = pairwise_merge_step(
        snippets=snippets,
        labels=labels,
        templates=templates,
        times=times,
        detect_sign=-1,
        unit_ids=np.unique(labels),
        detect_time_radius=10
    )
    assert len(new_times) == len(new_labels)
    assert len(new_times) <= len(times)

def test_get_times_labels_from_sorting():
    recording, sorting = se.toy_example(duration=6, num_channels=4, num_units=10, sampling_frequency=30000, seed=0, num_segments=1)
    recording: si.BaseRecording = recording
    sorting: si.BaseSorting = sorting

    times, labels = get_times_labels_from_sorting(sorting)
    assert len(times) == len(labels)
    assert len(np.unique(labels)) == sorting.get_num_units()

    # test the case of an empty sorting
    sorting2 = se.NumpySorting(sampling_frequency=30000)
    times2, labels2 = get_times_labels_from_sorting(sorting2)
    assert len(times2) == 0
    assert len(labels2) == 0


from typing import List, Tuple, Union, Dict
import numpy as np
import numpy.typing as npt
from dataclasses import dataclass
from sklearn import decomposition
from sklearn.neighbors import NearestNeighbors


class SnippetClassifier:
    def __init__(self, npca: int) -> None:
        self.npca = npca
        self.training_batches: List[TrainingBatch] = []
        self.pca_model = None
    def add_training_snippets(self, snippets: npt.NDArray[np.float32], label: int, offset: int):
        self.training_batches.append(TrainingBatch(snippets=snippets, label=label, offset=offset))
    def fit(self):
        if len(self.training_batches) == 0:
            raise Exception('No training batches added for classifier.') # pragma: no cover
        all_training_snippets = np.concatenate([b.snippets for b in self.training_batches], axis=0)
        L = all_training_snippets.shape[0]
        self.T = all_training_snippets.shape[1]
        self.M = all_training_snippets.shape[2]
        self.all_training_labels = np.concatenate([np.ones((b.num_snippets,), dtype=np.int32) * b.label for b in self.training_batches])
        self.all_training_offsets = np.concatenate([np.ones((b.num_snippets,), dtype=np.int32) * b.offset for b in self.training_batches])
        self.pca_model = decomposition.PCA(n_components=min(self.npca, L))
        self.pca_model.fit(all_training_snippets.reshape(L, self.T * self.M))
        X = self.pca_model.transform(all_training_snippets.reshape(L, self.T * self.M))
        self.nearest_neighbor_model = NearestNeighbors(n_neighbors=2)
        self.nearest_neighbor_model.fit(X)
    def classify_snippets(self, snippets: npt.NDArray[np.float32]) -> Tuple[Union[npt.NDArray[np.int32], None], Union[npt.NDArray[np.int32], None]]:
        if self.pca_model is None:
            raise Exception('self.pca_model is None, which probably means that fit() was not called.') # pragma: no cover
        Y = self.pca_model.transform(snippets.reshape(snippets.shape[0], self.T * self.M))
        nearest_inds = self.nearest_neighbor_model.kneighbors(Y, n_neighbors=2, return_distance=False)
        inds = nearest_inds[:, 1] # don't use the first because that could be an identical match
        return self.all_training_labels[inds], self.all_training_offsets[inds]
    def apply_label_mapping(self, mapping: Dict[int, int]):
        for k1, k2 in mapping.items():
            self.all_training_labels[self.all_training_labels == k1] = k2

@dataclass
class TrainingBatch:
    snippets: npt.NDArray[np.float32]
    label: int
    offset: int
    @property
    def num_snippets(self):
        return self.snippets.shape[0]


import numpy as np
import numpy.typing as npt


def compute_templates(snippets: npt.NDArray[np.float32], labels: np.int32):
    # L = snippets.shape[0]
    T = snippets.shape[1]
    M = snippets.shape[2]
    K = np.max(labels)
    templates = np.zeros((K, T, M), dtype=np.float32)
    for k in range(1, K + 1):
        snippets1 = snippets[labels == k]
        templates[k - 1] = np.median(snippets1, axis=0)
    return templates
        


from typing import Dict, Tuple, List, Union
import numpy as np
import numpy.typing as npt
import math
import spikeinterface as si
from .Scheme2SortingParameters import Scheme2SortingParameters
from .Scheme1SortingParameters import Scheme1SortingParameters
from ..core.detect_spikes import detect_spikes
from ..core.extract_snippets import extract_snippets, extract_snippets_in_channel_neighborhood
from .sorting_scheme1 import sorting_scheme1
from ..core.SnippetClassifier import SnippetClassifier
from ..core.pairwise_merge_step import remove_duplicate_events
from ..core.get_sampled_recording_for_training import get_sampled_recording_for_training
from ..core.get_times_labels_from_sorting import get_times_labels_from_sorting


def sorting_scheme2(
    recording: si.BaseRecording, *,
    sorting_parameters: Scheme2SortingParameters,
    return_snippet_classifiers: bool = False,
    reference_snippet_classifiers: Union[Dict[int, SnippetClassifier], None] = None,
    label_offset: int = 0
) -> Union[si.BaseSorting, Tuple[si.BaseSorting, Dict[int, SnippetClassifier]]]:
    """MountainSort 5 sorting scheme 2

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        sorting_parameters (Scheme2SortingParameters): Sorting parameters

    Returns:
        si.BaseSorting: SpikeInterface sorting object
    """

    if recording.get_num_segments() > 1:
        print('Recording has multiple segments. Joining segments for sorting...')
        recording_joined = si.concatenate_recordings(recording_list=[recording])
        sorting_joined, snippet_classifiers = sorting_scheme2(
            recording_joined,
            sorting_parameters=sorting_parameters,
            return_snippet_classifiers=True,
            reference_snippet_classifiers=reference_snippet_classifiers,
            label_offset=label_offset
        )
        print('Splitting sorting into segments to match original multisegment recording...')
        sorting = si.split_sorting(sorting_joined, recording_joined)
        if return_snippet_classifiers:
            return sorting, snippet_classifiers
        else:
            return sorting

    M = recording.get_num_channels()
    N = recording.get_num_frames()
    sampling_frequency = recording.sampling_frequency
    channel_locations = recording.get_channel_locations()

    sorting_parameters.check_valid(M=M, N=N, sampling_frequency=sampling_frequency, channel_locations=channel_locations)

    # Subsample the recording for training
    if sorting_parameters.training_duration_sec is not None:
        training_recording = get_sampled_recording_for_training(
            recording=recording,
            training_duration_sec=sorting_parameters.training_duration_sec,
            mode=sorting_parameters.training_recording_sampling_mode
        )
    else:
        training_recording = recording

    # Run the first phase of spike sorting (same as sorting_scheme1)
    sorting1 = sorting_scheme1(
        recording=training_recording,
        sorting_parameters=Scheme1SortingParameters(
            detect_threshold=sorting_parameters.phase1_detect_threshold,
            detect_sign=sorting_parameters.detect_sign,
            detect_time_radius_msec=sorting_parameters.phase1_detect_time_radius_msec,
            detect_channel_radius=sorting_parameters.phase1_detect_channel_radius,
            snippet_mask_radius=sorting_parameters.snippet_mask_radius,
            snippet_T1=sorting_parameters.snippet_T1,
            snippet_T2=sorting_parameters.snippet_T2,
            npca_per_branch=sorting_parameters.phase1_npca_per_branch,
            pairwise_merge_step=sorting_parameters.phase1_pairwise_merge_step
        )
    )
    times, labels = get_times_labels_from_sorting(sorting1)
    K = np.max(labels)
    labels = labels + label_offset

    print('Loading training traces')
    # Load the traces from the training recording
    training_traces = training_recording.get_traces()
    training_snippets = extract_snippets(
        traces=training_traces,
        channel_locations=None,
        mask_radius=None,
        times=times,
        channel_indices=None,
        T1=sorting_parameters.snippet_T1,
        T2=sorting_parameters.snippet_T2
    )

    print('Training classifier')
    # Train the classifier based on the labels obtained from the first phase sorting
    channel_masks: Dict[int, Union[List[int], None]] = {} # by channel
    for m in range(M):
        channel_masks[m] = []
        for m2 in range(M):
            if sorting_parameters.snippet_mask_radius is not None:
                if np.linalg.norm(channel_locations[m] - channel_locations[m2]) <= sorting_parameters.snippet_mask_radius:
                    channel_masks[m].append(m2)
            else:
                channel_masks[m] = None
    snippet_classifiers: Dict[int, SnippetClassifier] = {} # by channel
    for m in range(M):
        snippet_classifiers[m] = SnippetClassifier(npca=sorting_parameters.classifier_npca)
        # Add random snippets to classifier with label 0 (a noise cluster for classification)
        uniformly_spread_times = np.floor(np.linspace(sorting_parameters.snippet_T1, training_traces.shape[0] - sorting_parameters.snippet_T2 - 1, sorting_parameters.max_num_snippets_per_training_batch)).astype(np.int32)
        random_snippets = extract_snippets_in_channel_neighborhood(
            traces=training_traces,
            times=uniformly_spread_times,
            neighborhood=channel_masks[m],
            T1=sorting_parameters.snippet_T1,
            T2=sorting_parameters.snippet_T2
        )
        snippet_classifiers[m].add_training_snippets(random_snippets, label=0, offset=0)

    for k in range(label_offset + 1, label_offset + K + 1):
        inds0 = np.where(labels == k)[0]
        snippets0 = training_snippets[inds0]
        template0 = np.median(snippets0, axis=0) # T x M
        if sorting_parameters.detect_sign < 0:
            AA = -template0
        elif sorting_parameters.detect_sign > 0: # pragma: no cover
            AA = template0 # pragma: no cover
        else:
            AA = np.abs(template0) # pragma: no cover
        peak_indices_over_channels = np.argmax(AA, axis=0)
        peak_values_over_channels = np.max(AA, axis=0)
        peaks_to_include: List[dict] = []
        for m in range(M):
            if peak_values_over_channels[m] >= sorting_parameters.detect_threshold * 0.4: # should be a parameter
                peaks_to_include.append({
                    'channel': m,
                    'offset': peak_indices_over_channels[m] - sorting_parameters.snippet_T1
                })
        for peak in peaks_to_include:
            m = peak['channel']
            offset = peak['offset']
            if channel_masks[m] is not None:
                snippets0_masked = snippets0[:, :, channel_masks[m]]
            else:
                snippets0_masked = snippets0
            snippet_classifiers[m].add_training_snippets(
                snippets=subsample_snippets(np.roll(snippets0_masked, shift=-offset, axis=1), sorting_parameters.max_num_snippets_per_training_batch),
                label=k,
                offset=offset
            )
    training_snippets = None # Free up memory
    print('Fitting models')
    for m in range(M):
        snippet_classifiers[m].fit()

    # Now that we have the classifier, we can do the full sorting
    # Iterate over time chunks, detect and classify all spikes, and collect the results
    chunk_size = int(math.ceil(100e6 / recording.get_num_channels())) # size of chunks in samples
    print(f'Chunk size: {chunk_size / recording.sampling_frequency} sec')
    chunks = get_time_chunks(recording.get_num_samples(), chunk_size=chunk_size, padding=1000)
    times_list: list[npt.NDArray[np.int64]] = []
    labels_list: list[npt.NDArray[np.int32]] = []
    labels_reference_list = [] if reference_snippet_classifiers is not None else None
    for i, chunk in enumerate(chunks):
        print(f'Time chunk {i + 1} of {len(chunks)}')
        print('Loading traces')
        traces_chunk = recording.get_traces(start_frame=chunk.start - chunk.padding_left, end_frame=chunk.end + chunk.padding_right)
        print('Detecting spikes')
        time_radius = int(math.ceil(sorting_parameters.detect_time_radius_msec / 1000 * sampling_frequency))
        times_chunk, channel_indices_chunk = detect_spikes(
            traces=traces_chunk,
            channel_locations=channel_locations,
            time_radius=time_radius,
            channel_radius=sorting_parameters.detect_channel_radius,
            detect_threshold=sorting_parameters.detect_threshold,
            detect_sign=sorting_parameters.detect_sign,
            margin_left=sorting_parameters.snippet_T1,
            margin_right=sorting_parameters.snippet_T2,
            verbose=False
        )

        print('Extracting and classifying snippets')
        labels_chunk = np.zeros(len(times_chunk), dtype='int32')
        labels_reference_chunk = np.zeros(len(times_chunk), dtype='int32') if reference_snippet_classifiers is not None else None
        for m in range(M):
            inds = np.where(channel_indices_chunk == m)[0]
            if len(inds) > 0:
                snippets2 = extract_snippets_in_channel_neighborhood(
                    traces=traces_chunk,
                    times=times_chunk[inds],
                    neighborhood=channel_masks[m],
                    T1=sorting_parameters.snippet_T1,
                    T2=sorting_parameters.snippet_T2
                )
                labels_chunk_m, offsets_chunk_m = snippet_classifiers[m].classify_snippets(snippets2)
                if labels_chunk_m is not None:
                    labels_chunk[inds] = labels_chunk_m
                    times_chunk[inds] = times_chunk[inds] - offsets_chunk_m
                if reference_snippet_classifiers is not None:
                    labels_reference_chunk_m, _ = reference_snippet_classifiers[m].classify_snippets(snippets2)
                    if labels_reference_chunk_m is not None:
                        labels_reference_chunk[inds] = labels_reference_chunk_m

        # remove events with label 0
        valid_inds = np.where(labels_chunk > 0)[0]
        times_chunk: npt.NDArray[np.int32] = times_chunk[valid_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[valid_inds]
        labels_reference_chunk = labels_reference_chunk[valid_inds] if reference_snippet_classifiers is not None else None

        # now that we offset them we need to re-sort
        sort_inds2 = np.argsort(times_chunk)
        times_chunk: npt.NDArray[np.int32] = times_chunk[sort_inds2]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[sort_inds2]
        labels_reference_chunk = labels_reference_chunk[sort_inds2] if reference_snippet_classifiers is not None else None

        print('Removing duplicates')
        new_inds = remove_duplicate_events(times_chunk, labels_chunk, tol=time_radius)
        times_chunk: npt.NDArray[np.int32] = times_chunk[new_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[new_inds]
        labels_reference_chunk = labels_reference_chunk[new_inds] if reference_snippet_classifiers is not None else None

        # remove events in the margins
        valid_inds = np.where((chunk.padding_left <= times_chunk) & (times_chunk < chunk.total_size - chunk.padding_right))[0]
        times_chunk: npt.NDArray[np.int32] = times_chunk[valid_inds]
        labels_chunk: npt.NDArray[np.int32] = labels_chunk[valid_inds]
        labels_reference_chunk = labels_reference_chunk[valid_inds] if reference_snippet_classifiers is not None else None

        # don't forget to cast to int64 add the chunk start time
        times_list.append(
            times_chunk.astype(np.int64) + chunk.start - chunk.padding_left
        )
        labels_list.append(labels_chunk)
        if reference_snippet_classifiers is not None:
            labels_reference_list.append(labels_reference_chunk)

    # Now concatenate the results
    times_concat: npt.NDArray[np.int64] = np.concatenate(times_list)
    labels_concat: npt.NDArray[np.int32] = np.concatenate(labels_list)
    labels_reference_concat = np.concatenate(labels_reference_list) if reference_snippet_classifiers is not None else None

    if reference_snippet_classifiers is not None:
        mapping = get_labels_to_reference_labels_mapping(labels_concat, labels_reference_concat, label_offset=label_offset)
        print('==== mapping =======================')
        for k1, k2 in mapping.items():
            print(f'{k1} -> {k2}')
        print('====================================')
        for m in range(M):
            snippet_classifiers[m].apply_label_mapping(mapping)
        for k1, k2 in mapping.items():
            labels_concat[labels_concat == k1] = k2

    # Now create a new sorting object from the times and labels results
    sorting2 = si.NumpySorting.from_times_labels([times_concat], [labels_concat], sampling_frequency=recording.sampling_frequency)

    if return_snippet_classifiers:
        return sorting2, snippet_classifiers
    else:
        return sorting2

# Here's what this function does:
# 1. For each unit, find the matching unit in the reference (has to be a MUTUAL >0.5 overlap)
# 2. If the matching unit is found, then map the unit to the matching unit
# 3. If the matching unit is not found, then map it to the smallest unused label starting with label_offset+1
def get_labels_to_reference_labels_mapping(labels: npt.NDArray[np.int32], labels_reference: npt.NDArray[np.int32], *, label_offset) -> Dict[int, int]:
    mapping: Dict[int, int] = {}
    unique_labels = np.sort(np.unique(labels))
    last_used_k = label_offset
    for k in unique_labels:
        mapping[k] = None # initialize to None, if it stays as None, then we will need to create a new label
        inds = np.where(labels == k)[0]
        a = labels_reference[inds]
        k_refs, k_ref_counts = np.unique(a, return_counts=True)
        for ii in range(len(k_refs)):
            if k_ref_counts[ii] > 0.5 * len(inds): # the 0.5 is chosen so we don't map to the same unit twice
                inds_ref = np.where(labels_reference == k_refs[ii])[0] # import to test the other way around
                if k_ref_counts[ii] > 0.5 * len(inds_ref): # mutual overlap
                    mapping[k] = k_refs[ii] # map to the reference label
                    break
        if mapping[k] is None: # if not mapped to reference label, then create a new label
            mapping[k] = last_used_k + 1
            last_used_k = mapping[k]
    return mapping

class TimeChunk:
    def __init__(self, start: np.int64, end: np.int64, padding_left: np.int32, padding_right: np.int32):
        self.start = start
        self.end = end
        self.padding_left = padding_left
        self.padding_right = padding_right
        self.total_size = self.end - self.start + np.int64(padding_left) + np.int64(padding_right)

def get_time_chunks(num_samples: np.int64, chunk_size: np.int32, padding: np.int32) -> List[TimeChunk]:
    """Get time chunks
    Inputs:
        num_samples: number of samples in the recording
        chunk_size: size of each chunk in samples
        padding: padding on each side of the chunk in samples
    Returns:
        chunks: list of TimeChunk objects
    """
    chunks = []
    start = np.int64(0)
    while start < num_samples:
        end = np.int64(start) + np.int64(chunk_size)
        if end > num_samples:
            end = num_samples
        padding_left = min(padding, start)
        padding_right = min(padding, num_samples - end)
        chunks.append(TimeChunk(start=start, end=end, padding_left=padding_left, padding_right=padding_right))
        start = end
    return chunks

def subsample_snippets(snippets: npt.NDArray[np.float32], max_num: int) -> np.ndarray:
    """Subsample snippets
    Inputs:
        snippets: 3D array of snippets (num_snippets x T x M)
        max_num: maximum number of snippets to return
    Returns:
        snippets_out: 3D array of snippets (num_snippets x T x M)
    """
    num_snippets = snippets.shape[0]
    if num_snippets > max_num:
        inds = np.arange(0, max_num) * num_snippets // max_num
        snippets_out = snippets[inds]
    else:
        snippets_out = snippets
    return snippets_out


from setuptools import setup, find_packages

# read the contents of README.md
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

__version__ = '0.1.2'

setup(
    name='mountainsort5',
    version=__version__,
    author="Jeremy Magland",
    author_email="jmagland@flatironinstitute.org",
    url="https://github.com/magland/mountainsort5",
    description="MountainSort 5 spike sorting algorithm",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=[
        'spikeinterface>=0.97.1',
        'isosplit6>=0.1.0',
        'scikit-learn'
    ],
    tests_require=[
        "pytest",
        "pytest-cov"
    ]
)


from .schemes.sorting_scheme1 import sorting_scheme1
from .schemes.Scheme1SortingParameters import Scheme1SortingParameters
from .schemes.sorting_scheme2 import sorting_scheme2
from .schemes.Scheme2SortingParameters import Scheme2SortingParameters
from .schemes.sorting_scheme3 import sorting_scheme3
from .schemes.Scheme3SortingParameters import Scheme3SortingParameters


import spikeinterface as si
import numpy as np
from typing import Literal


def get_sampled_recording_for_training(
    recording: si.BaseRecording, *,
    training_duration_sec: float,
    mode: Literal['initial', 'uniform'] = 'initial'
) -> si.BaseRecording:
    """Get a sampled recording for the purpose of training

    Args:
        recording (si.BaseRecording): SpikeInterface recording object
        training_duration_sec (float): Duration of the training in seconds
        mode (str): 'initial' or 'uniform'

    Returns:
        si.BaseRecording: SpikeInterface recording object
    """
    if training_duration_sec * recording.sampling_frequency >= recording.get_num_frames():
        # if the training duration is longer than the recording, then just use the entire recording
        return recording
    if mode == 'initial':
        traces = recording.get_traces(start_frame=0, end_frame=int(training_duration_sec * recording.sampling_frequency))
    elif mode == 'uniform':
        # use chunks of 10 seconds
        chunk_size = int(recording.sampling_frequency * min(10, training_duration_sec))
        # the number of chunks depends on the training duration
        num_chunks = int(np.ceil(training_duration_sec * recording.sampling_frequency / chunk_size))
        chunk_sizes = [chunk_size for i in range(num_chunks)]
        chunk_sizes[-1] = int(training_duration_sec * recording.sampling_frequency - (num_chunks - 1) * chunk_size)
        if num_chunks == 1:
            # if only 1 chunk, then just use the initial chunk
            traces = recording.get_traces(start_frame=0, end_frame=int(training_duration_sec * recording.sampling_frequency))
        else:
            # the spacing between the chunks
            spacing = int((recording.get_num_frames() - np.sum(chunk_sizes)) / (num_chunks - 1))
            traces_list: list[np.ndarray] = []
            tt = 0
            for i in range(num_chunks):
                start_frame = tt
                end_frame = int(start_frame + chunk_sizes[i])
                traces_list.append(recording.get_traces(start_frame=start_frame, end_frame=end_frame))
                tt += int(chunk_sizes[i] + spacing)
            traces = np.concatenate(traces_list, axis=0)
    else:
        raise Exception('Invalid mode: ' + mode) # pragma: no cover
    
    rec = si.NumpyRecording(
        traces_list=[traces],
        sampling_frequency=recording.sampling_frequency,
        channel_ids=recording.get_channel_ids()
    )
    rec.set_channel_locations(recording.get_channel_locations())
    return rec